"""Module containing command line arguments configuration source."""

import re
import sys

from confeasy import SNAKE_CASE_REPLACE_PATTERN


TWO_OR_MORE_UNDERSCORES_PATTERN = re.compile(r"_{2,}")


class CommandLine:
    """Command line arguments configuration source"""

    def __init__(self, args: list[str] | None = None):
        self._args = sys.argv[1:] if args is None else args

    def get_configuration_data(self) -> dict[str, str | int | float | bool]:
        """
        Get data which should be merged into configuration.
        The keys should follow the required pattern - see documentation in developer.md.
        """
        result = {}
        i = 0
        while i < len(self._args):
            arg = self._args[i]

            # "--key=value" format
            if "=" in arg and arg.startswith("--"):
                key, value = arg.split("=", 1)
                key = _fmt_key(key.lstrip("-"))
                result[key] = _fmt_val(value)

            # "--key value" format
            elif arg.startswith("--"):
                key = _fmt_key(arg.lstrip("-"))
                if i + 1 < len(self._args) and not self._args[i + 1].startswith("--"):
                    result[key] = _fmt_val(self._args[i + 1])
                    i += 1
                else:
                    result[key] = True

            i += 1

        return result


def _fmt_key(s: str) -> str:
    sc = SNAKE_CASE_REPLACE_PATTERN.sub("_", s)
    return TWO_OR_MORE_UNDERSCORES_PATTERN.sub(".", sc).lower()


def _fmt_val(s: str) -> str | int | float | bool:
    if len(s) > 12:
        return s

    if s in ("true", "True", "TRUE"):
        return True

    if s.lstrip("-").isdigit():
        try:
            return int(s)
        except ValueError:
            pass

    if re.fullmatch(r"^-?\d*\.\d+$", s):
        try:
            return float(s)
        except ValueError:
            pass

    return s



"""Module containing environment variables configuration source."""

import os


class EnvironmentVariables:
    """Environment variables configuration source"""

    def __init__(self, prefix: str | None = None):
        self._prefix: str | None = prefix

    def get_configuration_data(self) -> dict[str, str | int | float | bool]:
        """
        Get data which should be merged into configuration.
        The keys should follow the required pattern - see documentation in developer.md.
        """

        def fmt(k: str) -> str:
            k = k if skip == 0 else k[skip:]
            return k.lower().lstrip("_").replace("__", ".")

        result: dict[str, str | int | float | bool] = {}
        skip = 0 if self._prefix is None else len(self._prefix)
        for key, value in os.environ.items():
            if not self._prefix:
                result[fmt(key)] = value
                continue
            if key.startswith(self._prefix):
                result[fmt(key)] = value

        return result



"""Base confeasy module."""

from __future__ import annotations

import re
from typing import Any, Protocol, TypeVar


__version__ = "1.0.0"

T = TypeVar("T")
VALID_KEY_PATTERN = re.compile(r"^[a-z]+(\.[a-z_]+)*$")
MAX_BIND_DEPTH = 5
SNAKE_CASE_REPLACE_PATTERN = re.compile(r"(?<!^)(?=[A-Z][a-z]|[A-Z](?=[A-Z][a-z]|$))")


class Source(Protocol):
    """Defines protocol for configuration sources."""

    def get_configuration_data(self) -> dict[str, str | int | float | bool]:
        """
        Get data which should be merged into configuration.
        The keys should follow the required pattern - see documentation in developer.md.
        """
        raise NotImplementedError("method must be implemented by classes that conform to the protocol.")


class Configuration:
    """
    Configuration merged from all the sources.
    It provides methods to retrieve individual values or to populate property values of a class instance.
    """

    def __init__(self, data: dict[str, str | int | float | bool]):
        self._data = data

    @property
    def data(self) -> dict[str, str | int | float | bool]:
        """Get all available configuration data."""
        return self._data

    def get_value(self, key: str) -> str | int | float | bool | None:
        """Get individual value from configuration by key."""
        return self._data.get(key, None)

    def bind(self, instance: T, prefix: str | None = None) -> T:
        """
        Populate property values of a class instance in-place (it does not create a copy).

        :param instance: a class instance, it should have either plain fields and/or properties including setters
        :param prefix: only configuration values with keys starting with this prefix will be used
        :return: the same instance of T that has been passed in
        """
        data = self._data if prefix is None else _filter_data(prefix, self._data)
        for key, value in data.items():
            _bind(instance, key, value, 0)

        return instance


def _filter_data(prefix: str, data: dict[str, str | int | float | bool]) -> dict[str, str | int | float | bool]:
    def new_key(k: str) -> str:
        idx = prefix_len + 1 if k[prefix_len] == "." else prefix_len
        return k[idx:]

    prefix_len = len(prefix)
    return {new_key(k): v for k, v in data.items() if k.startswith(prefix)}


def _bind(instance: T, path: str, value: str | int | float | bool, depth: int) -> bool:
    member_name, path_remainder = _expand_path(path)

    if not hasattr(instance, member_name):
        return False  # no field or property of such name has been found

    attr = getattr(type(instance), member_name, None)
    if isinstance(attr, property) and attr.fset is None:
        return False  # the property is read-only

    current_value = getattr(instance, member_name, None)
    if current_value is None:
        setattr(instance, member_name, value)
        return True

    if isinstance(current_value, str):
        setattr(instance, member_name, _force_str(value))
        return True

    if isinstance(current_value, int):
        setattr(instance, member_name, _force_int(value))
        return True

    if isinstance(current_value, float):
        setattr(instance, member_name, _force_float(value))
        return True

    if isinstance(current_value, bool):
        setattr(instance, member_name, _force_bool(value))
        return True

    if isinstance(current_value, list):
        setattr(instance, member_name, value)
        return True

    if depth >= MAX_BIND_DEPTH:
        return False

    return _bind(current_value, path_remainder, value, depth + 1)


def _force_str(v: Any) -> str:
    return v if isinstance(v, str) else str(v)


def _force_int(v: Any) -> int:
    return v if isinstance(v, int) else int(v)


def _force_float(v: Any) -> float:
    return v if isinstance(v, float) else float(v)


def _force_bool(v: Any) -> bool:
    return v if isinstance(v, bool) else bool(v)


def _expand_path(path: str) -> tuple[str, str]:
    if "." in path:
        before_dot, after_dot = path.split(".", 1)
        return before_dot, after_dot
    else:
        return path, ""


class Builder:
    """Defines which configuration sources will be included in building the final configuration data."""

    def __init__(self, drop_invalid_keys: bool = True):
        self._sources: list[Source] = []
        self._data: dict[str, str | int | float | bool] = {}
        self._drop_invalid_keys = drop_invalid_keys

    def add_source(self, source: Source) -> Builder:
        """Add new configuration source."""
        self._sources.append(source)
        return self

    def add_data(self, data: dict[str, Any]) -> Builder:
        """Add configuration data explicitly without a configuration source. This is useful for testing."""
        sanitized = _sanitize(data, self._drop_invalid_keys)
        self._data.update(sanitized)
        return self

    def build(self) -> Configuration:
        """Create final configuration from all the sources previously defined."""
        merged = self._data.copy()
        for source in self._sources:
            src = source.get_configuration_data()
            sanitized = _sanitize(src, self._drop_invalid_keys)
            merged.update(sanitized)
        return Configuration(merged)


def _sanitize(data: dict[str, Any], drop_invalid_keys: bool = True) -> dict[str, str | int | float | bool]:
    sanitized = {}
    for key, value in data.items():
        key = key.lower()
        if not VALID_KEY_PATTERN.match(key):
            if drop_invalid_keys:
                continue
            else:
                raise ValueError(f"invalid key '{key}'")

        if isinstance(value, str | int | float | bool):
            sanitized[key] = value
        else:
            sanitized[key] = str(value)

    return sanitized



"""Module containing TOML files configuration source."""

from __future__ import annotations

from pathlib import Path
import tomllib

from confeasy import SNAKE_CASE_REPLACE_PATTERN


class TomlFile:
    """TOML files configuration source."""

    def __init__(self, base_dir: str | None = None):
        """
        :param base_dir: unless absolute path is defined for a TOML file,
        this is where the relative paths will be looked for.
        If no base_dir is passed, then current working directory is assumed.
        """
        self._base_dir: Path = Path.cwd() if base_dir is None else Path(base_dir)
        self._files: list[tuple[str, bool]] = []

    def optional(self, path: str) -> TomlFile:
        """Define optional TOML file. If the path does not exist it is silently ignored."""
        self._files.append((path, False))
        return self

    def required(self, path: str) -> TomlFile:
        """
        Define required TOML file.

        :raises ValueError: if the path does not exist.
        """
        self._files.append((path, True))
        return self

    def get_configuration_data(self) -> dict[str, str | int | float | bool]:
        """
        Get data which should be merged into configuration.
        The keys should follow the required pattern - see documentation in developer.md.
        """
        result = {}
        for path_str, is_required in self._files:
            path = Path(path_str)
            path = path if path.is_absolute() else self._base_dir / path

            if not path.exists():
                if is_required:
                    raise ValueError(f"configuration file {path} does not exist")
                continue

            with open(path, "rb") as file:
                toml = tomllib.load(file)
                flat = _flatten_dict(toml)
                result.update(flat)

        return result


def _flatten_dict(d: dict, parent: str = "") -> dict[str, str | int | float | bool]:
    result: dict[str, str | int | float | bool] = {}
    for k, v in d.items():
        key = f"{parent}.{k}" if parent else k
        if isinstance(v, dict):
            result.update(_flatten_dict(v, key))
        else:
            key = SNAKE_CASE_REPLACE_PATTERN.sub("_", key).lower()
            result[key] = v
    return result



"""Module containing INI files configuration source."""

from __future__ import annotations

from configparser import ConfigParser
from pathlib import Path

from confeasy import SNAKE_CASE_REPLACE_PATTERN


class IniFile:
    """
    INI files configuration source.

    If you use $root as the name of a section in an INI file, the keys under the $root section will be treated
    as though they have no section at all (meaning the prefix "$root." will be stripped from their names).
    This approach emulates configurations in other sources where keys can exist at the root level.
    """

    def __init__(self, base_dir: str | None = None):
        """
        :param base_dir: unless absolute path is defined for a INI file,
        this is where the relative paths will be looked for.
        If no base_dir is passed, then current working directory is assumed.
        """
        self._base_dir: Path = Path.cwd() if base_dir is None else Path(base_dir)
        self._files: list[tuple[str, bool]] = []

    def optional(self, path: str) -> IniFile:
        """Define optional INI file. If the path does not exist it is silently ignored."""
        self._files.append((path, False))
        return self

    def required(self, path: str) -> IniFile:
        """
        Define required INI file.

        :raises ValueError: if the path does not exist.
        """
        self._files.append((path, True))
        return self

    def get_configuration_data(self) -> dict[str, str | int | float | bool]:
        """
        Get data which should be merged into configuration.
        The keys should follow the required pattern - see documentation in developer.md.
        """
        result: dict[str, str | int | float | bool] = {}
        for path_str, is_required in self._files:
            path = Path(path_str)
            path = path if path.is_absolute() else self._base_dir / path

            if not path.exists():
                if is_required:
                    raise ValueError(f"configuration file {path} does not exist")
                continue

            config = ConfigParser()
            # The following line allows the casing of the keys remain as in the original source (file).
            config.optionxform = str  # type: ignore[method-assign, assignment]
            config.read(path)

            for section in config.sections():
                for k, v in config.items(section):
                    key = f"{section}.{k}" if section != "$root" else k
                    key = SNAKE_CASE_REPLACE_PATTERN.sub("_", key).lower()
                    result[key] = v

        return result
